/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#include "m6809.h"
#include "pia.h"
#include "sound.h"
#include "logging.h"
#include "types.h"

#define SAMPLE_RATE 22050
#define CHANNELS 1
#define FORMAT AFMT_U8 /* AFMT_U8 */
#define FRAGMENTS 2
#define BUFFER_SIZE 512
#define SAMPLE_CYCLES (OSCILLATOR_RATE / SAMPLE_RATE)
//#define BUFFER_CYCLES ((OSCILLATOR_RATE * BUFFER_SIZE) / SAMPLE_RATE)
#define BUFFER_CYCLES ((OSCILLATOR_RATE / SAMPLE_RATE) * BUFFER_SIZE)

static int sound_fd;
static int channels = CHANNELS;
static int format;

static Uint32 sound_cycle_base;
static Uint8 *buffer;
static Uint8 *buffer_end;
static Uint8 *wrptr;
static Uint8 lastsample;
static Uint8 convbuf[BUFFER_SIZE * 4];

static void sound_send_buffer(void);

int sound_init(void) {
	int rate = SAMPLE_RATE;
	int buffer_size = BUFFER_SIZE;
	int num_fragments = FRAGMENTS;
	int tmp, fragment_param;
	char *device = "/dev/dsp";

	channels = CHANNELS;
	format = FORMAT;
	sound_fd = open(device, O_WRONLY);
	if (sound_fd == -1) {
		LOG_ERROR("Couldn't open audio device %s: %s!\n", device, strerror(errno));
		return 0;
	}
	fragment_param = 0;
	tmp = buffer_size;
	while (tmp > 1) {
		tmp >>= 1;
		fragment_param++;
	}
	fragment_param |= (num_fragments << 16);
	tmp = fragment_param;
	if (ioctl(sound_fd, SNDCTL_DSP_SETFRAGMENT, &fragment_param) == -1
			|| tmp != fragment_param) {
		LOG_ERROR("Couldn't set audio buffer size/fragmentation limit.\n");
		return 0;
	}
	channels--;
	tmp = channels;
	if (ioctl(sound_fd, SNDCTL_DSP_STEREO, &channels) == -1
			|| tmp != channels) {
		LOG_ERROR("Couldn't set desired (%d) audio channels.  Got %d.\n", tmp + 1, channels + 1);
		return 0;
	}
	channels++;
	if (ioctl(sound_fd, SNDCTL_DSP_SETFMT, &format) == -1) {
		LOG_ERROR("Couldn't set desired audio format.\n");
		return 0;
	}
	tmp = rate;
	if (ioctl(sound_fd, SNDCTL_DSP_SPEED, &rate) == -1
			|| tmp != rate) {
		LOG_ERROR("Couldn't set desired (%dHz) audio rate.  Got %dHz.\n", tmp, rate);
		return 0;
	}
	LOG_DEBUG(0,"Set up audio device at %dHz, %d channels.\nOSS buffering %d fragments of %d bytes.\n", rate, channels, num_fragments, buffer_size);
	if (format != FORMAT || channels != CHANNELS) {
		LOG_DEBUG(1,"Format needs conversion.\n");
	}
	PIA_1A.wrfunc[0] = sound_write_byte;
	buffer = (Uint8 *)malloc(BUFFER_SIZE);
	memset(buffer, 0x80, BUFFER_SIZE);
	buffer_end = buffer + BUFFER_SIZE;
	wrptr = buffer;
	lastsample = 0x00;
	sound_cycle_base = 0;
	return 1;
}

void sound_reset(void) {
}

void sound_write_byte(Uint32 octet) {
	Uint32 elapsed_cycles = current_cycle - sound_cycle_base;
	Uint32 fill_with;
	Uint8 *fill_to;
	if (elapsed_cycles > BUFFER_CYCLES * 2) {
		LOG_WARN("sound_cycle_base = %x\tcurrent_cycle = %x\t%x\n", sound_cycle_base, current_cycle, BUFFER_CYCLES);
		sound_cycle_base = current_cycle;
		elapsed_cycles = 0;
	}
	fill_to = buffer + (elapsed_cycles/(Uint32)SAMPLE_CYCLES);
	fill_with = lastsample;
	//octet >>= 1;
	//octet ^= 0x80;
	if (!(PIA_1B.control_register & 0x08) || PIA_0B.control_register & 0x08 || PIA_0A.control_register & 0x08)
		octet = lastsample; // 0x80;
	while (wrptr < fill_to) {
		while (wrptr < fill_to && wrptr < buffer_end)
			*(wrptr++) = fill_with;
		if (wrptr >= buffer_end) {
			sound_send_buffer();
			fill_to -= BUFFER_SIZE;
			sound_cycle_base += BUFFER_CYCLES;
			wrptr = buffer;
		}
	}
	lastsample = octet;
	return;
}

static void sound_send_buffer(void) {
	Uint8 *source = buffer;
	Uint8 *dest = convbuf;
	Uint8 tmp;
	Uint32 i;
	if (format == AFMT_U8 && channels == 1) {
		write(sound_fd, buffer, BUFFER_SIZE);
		return;
	}
	if (format == AFMT_U8 && channels == 2) {
		for (i = BUFFER_SIZE; i; i--) {
			tmp = *(source++);
			*(dest++) = tmp;
			*(dest++) = tmp;
		}
		write(sound_fd, convbuf, BUFFER_SIZE * 2);
		return;
	}
	if (format == AFMT_S16_LE) {
		for (i = BUFFER_SIZE; i; i--) {
			tmp = *(source++)^0x80;
			*(dest++) = 0;
			*(dest++) = tmp;
			if (channels == 2) {
				*(dest++) = 0;
				*(dest++) = tmp;
			}
		}
		write(sound_fd, convbuf, BUFFER_SIZE * channels * 2);
		return;
	}
	if (format == AFMT_S16_BE) {
		for (i = BUFFER_SIZE; i; i--) {
			tmp = *(source++)^0x80;
			*(dest++) = tmp;
			*(dest++) = 0;
			if (channels == 2) {
				*(dest++) = tmp;
				*(dest++) = 0;
			}
		}
		write(sound_fd, convbuf, BUFFER_SIZE * channels * 2);
		return;
	}
}
